/**
 * @Author: wzy
 * @Date: 2021-10-25
 * @LastEditTime: 
 * @LastEditors: Please set LastEditors
 * @Description: In User Settings Edit
 * @filePath:
 */

#pragma once

#include <iostream>
#include <algorithm>
#include <vector>
#include <map>
#include "glog/logging.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>

// #include "mpc_drv_msgs/msg/object.hpp"
#include "PD_EgoVehStsArb.h"
#include "PD_SVEstmd.h"
#include "PD_commonfunc.h"
// #include "../../../../include/CtApBehaviorAnalysis_/interface/rm_road_info.h"
#include "dpc/impl_type_roadinfo.h"

#define FLAG_OUTPUT 0         // switch 1
#define FLAG_DLOG_IOOUT 0     // switch 1
#define FLAG_OFS_IOOUT 0      // switch 1
#if (FLAG_OFS_IOOUT != 0)||(FLAG_DLOG_IOOUT !=0)||(FLAG_OUTPUT != 0)
#include <fstream>
#endif

namespace ADC
{
namespace PD_DynmcObjRsk
{

/**
 * @brief 
 * @param 
 * @return 
 */

struct PDEvaVar_
{
  uint32_t val_obj_id;
  uint8_t val_ttc = 0;
  uint8_t val_td = 0;
  float val_lat_dist = 0.0f;
  float val_lgt_dist = 0.0f;
};

// ego vehicle trajectory point
struct EgoVeh_Point
{
    float t;           //时间
    float s;           //纵向位移
    float l;           //横向位移
    float vs;          //纵向速度
    float as;          //纵向加速度
    float jks;         //纵向加加速度
    EgoVeh_Point()
    {
        t = 0.0f;
        s = 0.0f;
        l = 0.0f;
        vs = 0.0f;
        as = 0.0f;
        jks = 0.0f;
    }
};

// surrounding vehicle trajectory point
struct SV_Point
{
  float t;                 //时间
  float s_rel;             //相对自车的纵向位移
  float l;                 //横向位移
  float vs;                //纵向速度
  float vl;                //横向速度
  SV_Point()
  {
    t = 0.0f;
    s_rel = 0.0f;
    l = 0.0f;
    vs = 0.0f;
    vl = 0.0f;
  }
};
struct SurdVeh_TrjPnt
{
  uint32_t obj_ID;                       //环境车ID
  uint8_t surdVeh_Class;                 //车辆类型
  float surdVeh_Width;                   //车辆宽度
  float surdVeh_Length;                  //车辆长度
  std::vector<SV_Point> sv_point_list;   //环境车轨迹序列
};
// Dynamic risk
struct DynamicRisk
{
  std::vector<float> m_ObjLongPreClsnTTC;             // 纵向预碰撞时间TTC
  std::vector<float> m_ObjLongPreClsnTD;              // 纵向预碰撞时距TD
  std::vector<float> m_ObjLongOvlpDeg;                // 纵向重叠度
  std::vector<float> m_ObjLatOvlpDeg;                 // 横向重叠度
  float m_ObjLongMinDis = 0.0f;                       // 纵向最小距离
  float m_ObjLatMinDis = 0.0f;                        // 横向最小距离 
  float m_ObjLongCrsSfZnItg = 0.0f;                   // 纵向超安全区域积分
  float m_ObjLatCrsSfZnItg = 0.0f;                    // 横向超安全区域积分
  uint32_t ObjID;                                     // 环境车ID
};

struct TmpDynamicRisk
{
  std::vector<float> tmp_ObjLongPreClsnTTC;          
  std::vector<float> tmp_ObjLongPreClsnTD;            
  std::vector<float> tmp_ObjLongOvlpDeg;             
  std::vector<float> tmp_ObjLatOvlpDeg;             
  std::vector<float> tmp_ObjLongMinDis;              
  std::vector<float> tmp_ObjLatMinDis;            
  // std::vector<float> tmp_ObjLongCrsSfZnItg;       
  // std::vector<float> tmp_ObjLatCrsSfZnItg;        
  uint32_t tmp_ObjID;                                    
};

class DynamicObjRisk
{
public:
  DynamicObjRisk(const PD::PD_Input &pd_input,
                 const std::vector<ADC::PD_EgoVehTrj::EgoVeh_TrjPnt>& EgoVeh_trace_list,
                 const std::vector<ADC::PD_SVEstmd::PD_obj_SV>& SudVeh_trace_list);
  ~DynamicObjRisk() = default;

private:
  bool init(const PD::PD_Input &pd_input,
            const std::vector<ADC::PD_EgoVehTrj::EgoVeh_TrjPnt>& EgoVeh_trace_list,
            const std::vector<ADC::PD_SVEstmd::PD_obj_SV>& SudVeh_trace_list);
  bool initFlag = true;

private:
  // ego vehicle trajectory information
  std::vector<EgoVeh_Point> _egoVeh_TrjPnt;
  // one surrounding vehicle trajectory information
  std::vector<SV_Point> _sv_point_list;
  // allsurrounding vehicle trajectory information
  std::vector<SurdVeh_TrjPnt> _allSurdVeh_TrjPnt;
  std::vector<Object> _object_list;
  std::vector<float> _latthreshold_list;
  std::vector<bool> _threshold_boundary_flag;  
  std::vector<bool> _LgOvlpFlag;
  std::vector<bool> _LgMinDisFlag;
  float _timesampl = 0.1f;

  bool FlToVctFlag = false;
  uint8_t _ego_lane_id = 0;
  uint32_t _ego_front_id = 0;
  uint32_t _Target_front_id = 0;
  uint32_t _Target_back_id = 0;
  uint32_t _ego_back_id = 0;
  std::vector<uint32_t> interest_obj_id_list;
  std::vector<Lane> _lane_list;
  std::map<uint32_t, TmpDynamicRisk> _tmpdynamicrisk;
  std::vector<ObjectLocation> ObjInEgoLaneList;
  std::vector<ObjectLocation> ObjInEgoLaneListL;
  std::vector<ObjectLocation> ObjInEgoLaneListLL;
  std::vector<ObjectLocation> ObjInEgoLaneListR;
  std::vector<ObjectLocation> ObjInEgoLaneListRR;
  std::vector<ObjectLocation> ObjListOther;

  struct Index_
  {
    int x = 0;
    int y = 0;
  };
  Index_ RoadID;
  Index_ PassageID;
  uint8_t Action_Type = 0;

  #define VEHCALPRMVEHLEN 4.895   
  #define VEHCALPRMVEHWIDTH 1.926   
  #define VEHCALPRMFORAXLEDSTRETOVEHFRNT 3.874  
  #define VEHCALPRMFORREARAXISTOREARBUMPER 1.021

  //Carinfo
  // float _egoVeh_Width = 1.9f;
  //   //relative position of rear axle center
  // float _egoVeh_Length = 3.8f;
  uint8_t _surdVeh_Class = 0;
  float _surdVeh_Width = 0.0f;
  float _surdVeh_Length = 0.0f;

public:
  //output
  void process();
  std::vector<DynamicRisk> m_DynObjRsk_list;       // 交通参与者导致的风险
  std::vector<PDEvaVar_> pd_eva_val_;        
  
private:
  void ObserveVal();
  void OutPutToOfs();
  std::vector<bool> ThresholdBoundary();
  // void ThresholdBoundary();

  // std::vector<DynamicRisk> DynamicRiskPostProcess(const uint8_t &_ego_lane_id,
  //                                                 const std::vector<Lane> &_lane_list,
  //                                                 std::map<uint32_t, TmpDynamicRisk> &_tmpdynamicrisk);
  std::vector<DynamicRisk> DynamicRiskPostProcess(std::map<uint32_t, TmpDynamicRisk> &_tmpdynamicrisk);

  void CalculateDynamicRisk();
  void DynamicRiskDefultOut();
  bool ObjctOpt();
  float CalLgMinDis(std::vector<float> &LgDisList);
  float CalLatMinDis(std::vector<float> &LatDisList);
  void PushFun(std::map<uint32_t, TmpDynamicRisk>::iterator &pos, const int &lengsize);
  void GetNeighborLaneObjID(const uint8_t &_ego_lane_id,
                            const std::vector<Lane> &_lane_id);
  // std::vector<float> operator()(std::vector<float>::iterator &it, std::vector<float>::iterator &it + int Dis);

  template <class T1>
  T1 InterPoint(const std::vector<T1> &x_, const std::vector<T1> &y_, T1 &x_point);

/**
  * @brief   Calculate lateral threshold
  * @param   
  * @return  the vector of lateral threshold
  */
  std::vector<float> CalculateLatThreshold();

/**
  * @brief   Calculate TTC
  * @param   
  * @return  the vector of TTC
  */
  std::vector<float> CalculateTTC();

/**
  * @brief   Calculate TimeDistance
  * @param   
  * @return  the vector of TimeDistance
  */
  std::vector<float> CalculateTD();

/**
  * @brief   Calculate longitudinal overlap
  * @param   
  * @return  the vector of longitudinal overlap
  */
  std::vector<float> CalculateLongOvlpDeg();
 
/**
  * @brief   Calculate lateral overlap
  * @param   
  * @return  the vector of lateral overlap
  */
  std::vector<float> CalculateLatOvlpDeg();

/**
  * @brief   Calculate the minidistance of longitudinal
  * @param   
  * @return  the minidistance of longitudinal
  */
  std::vector<float> CalculateLongMinDis();
    
/**
  * @brief   Calculate the minidistance of lateral 
  * @param   
  * @return  the minidistance of lateral
  */
  std::vector<float> CalculateLatMinDis();

/**
  * @brief   Calculate the integral of cross safe zoon along longitudinal
  * @param   
  * @return  the integral of cross safe zoon along longitudinal
  */
  float CalculateLongCrsSfZnItg(std::vector<float> &LongCrsSfZn_list);

/**
  * @brief   Calculate the integral of cross safe zoon along lateral
  * @param   
  * @return  the integral of cross safe zoon along lateral
  */
  float CalculateLatCrsSfZnItg(std::vector<float> &LatCrsSfZn_list);

};

}    // namespace PD_DynmcObjRsk
}    //namespace ADC
